# -*- coding: utf-8 -*-
import contextlib
import logging
import os
import os.path

import yaml
try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:  # pragma: no cover
    from yaml import SafeLoader  # type: ignore
from bravado_core.spec import is_yaml
from six import iteritems
from six import itervalues
from six.moves import urllib
from six.moves.urllib import parse as urlparse

import simplejson
from bravado.requests_client import RequestsClient

log = logging.getLogger(__name__)


def is_file_scheme_uri(url):
    return urlparse.urlparse(url).scheme == u'file'


class FileEventual(object):
    """Adaptor which supports the :class:`crochet.EventualResult`
    interface for retrieving api docs from a local file.
    """

    class FileResponse(object):

        def __init__(self, data):
            self.text = data
            self.headers = {}

        def json(self):
            return simplejson.loads(self.text.decode('utf-8'))

    def __init__(self, path):
        self.path = path
        self.is_yaml = is_yaml(path)

    def get_path(self):
        if not self.path.endswith('.json') and not self.is_yaml:
            return self.path + '.json'
        return self.path

    def wait(self, **kwargs):
        with contextlib.closing(urllib.request.urlopen(self.get_path())) as fp:
            content = fp.read()
            return self.FileResponse(content)

    def result(self, *args, **kwargs):
        return self.wait(*args, **kwargs)

    def cancel(self):
        pass


def request(http_client, url, headers):
    """Download and parse JSON from a URL.

    :param http_client: a :class:`bravado.http_client.HttpClient`
    :param url: url for api docs
    :return: an object with a :func`wait` method which returns the api docs
    """
    if is_file_scheme_uri(url):
        return FileEventual(url)

    request_params = {
        'method': 'GET',
        'url': url,
        'headers': headers,
    }

    return http_client.request(request_params)


class Loader(object):
    """Abstraction for loading Swagger API's.

    :param http_client: HTTP client interface.
    :type  http_client: http_client.HttpClient
    :param request_headers: dict of request headers
    """

    def __init__(self, http_client, request_headers=None):
        self.http_client = http_client
        self.request_headers = request_headers or {}

    def load_spec(self, spec_url, base_url=None):
        """Load a Swagger Spec from the given URL

        :param spec_url: URL to swagger.json
        :param base_url: TODO: need this?
        :returns: json spec in dict form
        """
        response = request(
            self.http_client,
            spec_url,
            self.request_headers,
        ).result()

        content_type = response.headers.get('content-type', '').lower()
        if is_yaml(spec_url, content_type):
            return self.load_yaml(response.text)
        else:
            return response.json()

    def load_yaml(self, text):
        """Load a YAML Swagger spec from the given string, transforming
        integer response status codes to strings. This is to keep
        compatibility with the existing YAML spec examples in
        https://github.com/OAI/OpenAPI-Specification/tree/master/examples/v2.0/yaml
        :param text: String from which to parse the YAML.
        :type  text: basestring
        :return: Python dictionary representing the spec.
        :raise: yaml.parser.ParserError: If the text is not valid YAML.
        """
        data = yaml.load(text, Loader=SafeLoader)
        for methods in itervalues(data.get('paths', {})):
            for operation in itervalues(methods):
                if 'responses' in operation:
                    operation['responses'] = {
                        str(code): response
                        for code, response in iteritems(operation['responses'])
                    }

        return data


# TODO: Adding the file scheme here just adds complexity to request()
# Is there a better way to handle this?
def load_file(spec_file, http_client=None):
    """Loads a spec file

    :param spec_file: Path to swagger.json.
    :param http_client: HTTP client interface.
    :return: validated json spec in dict form
    :raise: IOError: On error reading swagger.json.
    """
    file_path = os.path.abspath(spec_file)
    url = urlparse.urljoin(u'file:', urllib.request.pathname2url(file_path))
    # When loading from files, everything is relative to the spec file
    dir_path = os.path.dirname(file_path)
    base_url = urlparse.urljoin(u'file:', urllib.request.pathname2url(dir_path))
    return load_url(url, http_client=http_client, base_url=base_url)


def load_url(spec_url, http_client=None, base_url=None):
    """Loads a Swagger spec.

    :param spec_url: URL for swagger.json.
    :param http_client: HTTP client interface.
    :param base_url:    Optional URL to be the base URL for finding API
                        declarations. If not specified, 'basePath' from the
                        resource listing is used.
    :return: validated spec in dict form
    :raise: IOError, URLError: On error reading api-docs.
    """
    if http_client is None:
        http_client = RequestsClient()

    loader = Loader(http_client=http_client)
    return loader.load_spec(spec_url, base_url=base_url)



# -*- coding: utf-8 -*-
"""
The :class:`SwaggerClient` provides an interface for making API calls based on
a swagger spec, and returns responses of python objects which build from the
API response.

Structure Diagram::

        +---------------------+
        |                     |
        |    SwaggerClient    |
        |                     |
        +------+--------------+
               |
               |  has many
               |
        +------v--------------+
        |                     |
        |     Resource        +------------------+
        |                     |                  |
        +------+--------------+         has many |
               |                                 |
               |  has many                       |
               |                                 |
        +------v--------------+           +------v--------------+
        |                     |           |                     |
        |     Operation       |           |    SwaggerModel     |
        |                     |           |                     |
        +------+--------------+           +---------------------+
               |
               |  uses
               |
        +------v--------------+
        |                     |
        |     HttpClient      |
        |                     |
        +---------------------+


To get a client

.. code-block:: python

    client = bravado.client.SwaggerClient.from_url(swagger_spec_url)
"""
import logging

from bravado_core.docstring import create_operation_docstring
from bravado_core.exception import SwaggerMappingError
from bravado_core.formatter import SwaggerFormat  # noqa
from bravado_core.param import marshal_param
from bravado_core.spec import Spec
from six import iteritems
from six import itervalues

from bravado.config import bravado_config_from_config_dict
from bravado.config import RequestConfig
from bravado.docstring_property import docstring_property
from bravado.requests_client import RequestsClient
from bravado.swagger_model import Loader
from bravado.warning import warn_for_deprecated_op

log = logging.getLogger(__name__)


class SwaggerClient(object):
    """A client for accessing a Swagger-documented RESTful service.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    """

    def __init__(self, swagger_spec, also_return_response=False):
        self.__also_return_response = also_return_response
        self.swagger_spec = swagger_spec

    @classmethod
    def from_url(cls, spec_url, http_client=None, request_headers=None, config=None):
        """Build a :class:`SwaggerClient` from a url to the Swagger
        specification for a RESTful API.

        :param spec_url: url pointing at the swagger API specification
        :type spec_url: str
        :param http_client: an HTTP client used to perform requests
        :type  http_client: :class:`bravado.http_client.HttpClient`
        :param request_headers: Headers to pass with http requests
        :type  request_headers: dict
        :param config: Config dict for bravado and bravado_core.
            See CONFIG_DEFAULTS in :module:`bravado_core.spec`.
            See CONFIG_DEFAULTS in :module:`bravado.client`.

        :rtype: :class:`bravado_core.spec.Spec`
        """
        log.debug(u"Loading from %s", spec_url)
        http_client = http_client or RequestsClient()
        loader = Loader(http_client, request_headers=request_headers)
        spec_dict = loader.load_spec(spec_url)

        # RefResolver may have to download additional json files (remote refs)
        # via http. Wrap http_client's request() so that request headers are
        # passed along with the request transparently. Yeah, this is not ideal,
        # but since RefResolver has new found responsibilities, it is
        # functional.
        if request_headers is not None:
            http_client.request = inject_headers_for_remote_refs(
                http_client.request, request_headers)

        return cls.from_spec(spec_dict, spec_url, http_client, config)

    @classmethod
    def from_spec(cls, spec_dict, origin_url=None, http_client=None,
                  config=None):
        """
        Build a :class:`SwaggerClient` from a Swagger spec in dict form.

        :param spec_dict: a dict with a Swagger spec in json-like form
        :param origin_url: the url used to retrieve the spec_dict
        :type  origin_url: str
        :param config: Configuration dict - see spec.CONFIG_DEFAULTS

        :rtype: :class:`bravado_core.spec.Spec`
        """
        http_client = http_client or RequestsClient()
        config = config or {}

        # Apply bravado config defaults
        bravado_config = bravado_config_from_config_dict(config)
        # remove bravado configs from config dict
        for key in set(bravado_config._fields).intersection(set(config)):
            del config[key]
        # set bravado config object
        config['bravado'] = bravado_config

        swagger_spec = Spec.from_dict(
            spec_dict, origin_url, http_client, config,
        )
        return cls(swagger_spec, also_return_response=bravado_config.also_return_response)

    def get_model(self, model_name):
        return self.swagger_spec.definitions[model_name]

    def _get_resource(self, item):
        """
        :param item: name of the resource to return
        :return: :class:`Resource`
        """
        resource = self.swagger_spec.resources.get(item)
        if not resource:
            raise AttributeError(
                'Resource {0} not found. Available resources: {1}'
                .format(item, ', '.join(dir(self))))

        # Wrap bravado-core's Resource and Operation objects in order to
        # execute a service call via the http_client.
        return ResourceDecorator(resource, self.__also_return_response)

    def __repr__(self):
        return u"%s(%s)" % (self.__class__.__name__, self.swagger_spec.api_url)

    def __getattr__(self, item):
        return self._get_resource(item)

    def __dir__(self):
        return self.swagger_spec.resources.keys()


def inject_headers_for_remote_refs(request_callable, request_headers):
    """Inject request_headers only when the request is to retrieve the
    remote refs in the swagger spec (vs being a request for a service call).

    :param request_callable: method on http_client to make a http request
    :param request_headers: headers to inject when retrieving remote refs
    """
    def request_wrapper(request_params, *args, **kwargs):

        def is_remote_ref_request(request_kwargs):
            # operation is only present for service calls
            return request_kwargs.get('operation') is None

        if is_remote_ref_request(kwargs):
            request_params['headers'] = request_headers

        return request_callable(request_params, *args, **kwargs)

    return request_wrapper


class ResourceDecorator(object):
    """
    Wraps :class:`bravado_core.resource.Resource` so that accesses to contained
    operations can be instrumented.
    """

    def __init__(self, resource, also_return_response=False):
        """
        :type resource: :class:`bravado_core.resource.Resource`
        """
        self.also_return_response = also_return_response
        self.resource = resource

    def __getattr__(self, name):
        """
        :rtype: :class:`CallableOperation`
        """
        return CallableOperation(getattr(self.resource, name), self.also_return_response)

    def __dir__(self):
        """
        Exposes correct attrs on resource when tab completing in a REPL
        """
        return self.resource.__dir__()


class CallableOperation(object):
    """Wraps an operation to make it callable and provides a docstring. Calling
    the operation uses the configured http_client.

    :type operation: :class:`bravado_core.operation.Operation`
    """

    def __init__(self, operation, also_return_response=False):
        self.also_return_response = also_return_response
        self.operation = operation

    @docstring_property(__doc__)
    def __doc__(self):
        return create_operation_docstring(self.operation)

    def __getattr__(self, name):
        """Forward requests for attrs not found on this decorator to the
        delegate.
        """
        return getattr(self.operation, name)

    def __call__(self, **op_kwargs):
        """Invoke the actual HTTP request and return a future.

        :rtype: :class:`bravado.http_future.HTTPFuture`
        """
        log.debug(u'%s(%s)', self.operation.operation_id, op_kwargs)
        warn_for_deprecated_op(self.operation)

        # Get per-request config
        request_options = op_kwargs.pop('_request_options', {})
        request_config = RequestConfig(request_options, self.also_return_response)

        request_params = construct_request(
            self.operation, request_options, **op_kwargs)

        http_client = self.operation.swagger_spec.http_client

        return http_client.request(
            request_params,
            operation=self.operation,
            request_config=request_config,
        )


def construct_request(operation, request_options, **op_kwargs):
    """Construct the outgoing request dict.

    :type operation: :class:`bravado_core.operation.Operation`
    :param request_options: _request_options passed into the operation
        invocation.
    :param op_kwargs: parameter name/value pairs to passed to the
        invocation of the operation.

    :return: request in dict form
    """
    url = operation.swagger_spec.api_url.rstrip('/') + operation.path_name
    request = {
        'method': str(operation.http_method.upper()),
        'url': url,
        'params': {},  # filled in downstream
        'headers': request_options.get('headers', {}),
    }
    # Adds Accept header to request for msgpack response if specified
    if request_options.get('use_msgpack', False):
        request['headers']['Accept'] = 'application/msgpack'

    # Copy over optional request options
    for request_option in ('connect_timeout', 'timeout'):
        if request_option in request_options:
            request[request_option] = request_options[request_option]

    construct_params(operation, request, op_kwargs)

    return request


def construct_params(operation, request, op_kwargs):
    """Given the parameters passed to the operation invocation, validates and
    marshals the parameters into the provided request dict.

    :type operation: :class:`bravado_core.operation.Operation`
    :type request: dict
    :param op_kwargs: the kwargs passed to the operation invocation

    :raises: SwaggerMappingError on extra parameters or when a required
        parameter is not supplied.
    """
    current_params = operation.params.copy()
    for param_name, param_value in iteritems(op_kwargs):
        param = current_params.pop(param_name, None)
        if param is None:
            raise SwaggerMappingError(
                "{0} does not have parameter {1}"
                .format(operation.operation_id, param_name))
        marshal_param(param, param_value, request)

    # Check required params and non-required params with a 'default' value
    for remaining_param in itervalues(current_params):
        if remaining_param.location == 'header' and remaining_param.name in request['headers']:
            marshal_param(remaining_param, request['headers'][remaining_param.name], request)
        else:
            if remaining_param.required:
                raise SwaggerMappingError(
                    '{0} is a required parameter'.format(remaining_param.name))
            if not remaining_param.required and remaining_param.has_default():
                marshal_param(remaining_param, None, request)



# -*- coding: utf-8 -*-
import logging
import sys

import crochet
import fido
import fido.exceptions
import requests
import requests.structures
import six
import twisted.internet.error
import twisted.web.client
import typing
from bravado_core.operation import Operation
from bravado_core.response import IncomingResponse
from yelp_bytes import to_bytes

from bravado.config import RequestConfig
from bravado.http_client import HttpClient
from bravado.http_future import FutureAdapter
from bravado.http_future import HttpFuture

if getattr(typing, 'TYPE_CHECKING', False):
    class _FidoStub(typing.Protocol):
        code = None  # type: int
        body = None  # type: str
        reason = None  # type: typing.Text
        headers = None  # type: typing.Mapping[bytes, typing.List[bytes]]

        def json(self, *args, **kwargs):
            # type: (typing.Any, typing.Any) -> typing.Mapping[typing.Text, typing.Any]
            pass


log = logging.getLogger(__name__)
T = typing.TypeVar('T')


class FidoResponseAdapter(IncomingResponse):
    """Wraps a fido.fido.Response object to provide a uniform interface
    to the response innards.

    :type fido_response: :class:`fido.fido.Response`
    """

    def __init__(self, fido_response):
        # type: (_FidoStub) -> None
        self._delegate = fido_response
        self._headers = None  # type: typing.Optional[typing.MutableMapping[typing.Text, typing.Text]]

    @property
    def status_code(self):
        # type: () -> int
        return self._delegate.code

    @property
    def text(self):
        # type: () -> typing.Text
        return self._delegate.body

    @property
    def raw_bytes(self):
        # type: () -> bytes
        return self._delegate.body

    @property
    def reason(self):
        # type: () -> typing.Text
        return self._delegate.reason

    @property
    def headers(self):
        # type: () -> typing.Mapping[typing.Text, typing.Text]
        # Header names and values are bytestrings, which is an issue on Python 3. Additionally,
        # header values are lists of strings. This is incompatible with how requests returns headers.
        # Let's match the requests interface so code dealing with headers continues to work even when
        # you change the HTTP client.
        if not self._headers:
            # using typing.cast here breaks on Python 3.5.1 and 3.5.0
            self._headers = requests.structures.CaseInsensitiveDict()  # type: ignore
            for header, values in self._delegate.headers.items():
                # header names are encoded using latin1, while header values are encoded using UTF-8.
                # We'll take the last entry in the list of values, making sure the latest header sent
                # takes precedence. The fact that twisted uses lists of strings for values seems to be
                # an edge case, I couldn't find any documentation or test using more than one entry in
                # the list of values for a given header.
                self._headers[header.decode('latin1')] = values[-1].decode('utf8')  # type: ignore

        return self._headers  # type: ignore

    def json(self, **_):
        # type: (typing.Any) -> typing.Mapping[typing.Text, typing.Any]
        # TODO: pass the kwargs downstream
        return self._delegate.json()


class FidoClient(HttpClient):
    """Fido (Asynchronous) HTTP client implementation.
    """

    def request(
        self,
        request_params,  # type: typing.MutableMapping[str, typing.Any]
        operation=None,  # type: typing.Optional[Operation]
        request_config=None,  # type: typing.Optional[RequestConfig]
    ):
        # type: (...) -> HttpFuture[T]
        """Sets up the request params as per Twisted Agent needs.
        Sets up crochet and triggers the API request in background

        :param request_params: request parameters for the http request.
        :type request_params: dict
        :param operation: operation that this http request is for. Defaults
            to None - in which case, we're obviously just retrieving a Swagger
            Spec.
        :type operation: :class:`bravado_core.operation.Operation`
        :param RequestConfig request_config: per-request configuration

        :rtype: :class: `bravado_core.http_future.HttpFuture`
        """

        request_for_twisted = self.prepare_request_for_twisted(request_params)

        future_adapter = FidoFutureAdapter(fido.fetch(**request_for_twisted))  # type: FidoFutureAdapter[T]

        return HttpFuture(future_adapter,
                          FidoResponseAdapter,
                          operation,
                          request_config)

    @staticmethod
    def prepare_request_for_twisted(request_params):
        # type: (typing.MutableMapping[str, typing.Any]) -> typing.Mapping[str, typing.Any]
        """
        Uses the python package 'requests' to prepare the data as per twisted
        needs. requests.PreparedRequest.prepare is able to compute the body and
        the headers for the http call based on the input request_params. This
        contains any query parameters, files, body and headers to include.

        :return: dictionary in the form
            {
                'body': string,  # (can represent any content-type i.e. json,
                    file, multipart..),
                'headers': dictionary,  # headers->values
                'method': string,  # can be 'GET', 'POST' etc.
                'url': string,
                'timeout': float,  # optional
                'connect_timeout': float,  # optional
            }
        """

        prepared_request = requests.PreparedRequest()

        # Ensure that all the headers are converted to strings.
        # This is need to workaround https://github.com/requests/requests/issues/3491
        request_params['headers'] = {
            k: v if isinstance(v, six.binary_type) else str(v)
            for k, v in six.iteritems(request_params.get('headers', {}))
        }

        prepared_request.prepare(
            headers=request_params.get('headers'),
            data=request_params.get('data'),
            params=request_params.get('params'),
            files=request_params.get('files'),
            url=request_params.get('url'),
            method=request_params.get('method')
        )

        # content-length was computed by 'requests' based on the current body
        # but body will be processed by fido using twisted FileBodyProducer
        # causing content-length to lose meaning and break the client.
        prepared_request.headers.pop('Content-Length', None)

        request_for_twisted = {
            # converting to string for `requests` method is necessary when
            # using requests < 2.8.1 due to a bug while handling unicode values
            # See changelog 2.8.1 at https://pypi.python.org/pypi/requests
            'method': str(prepared_request.method or 'GET'),
            'body': (
                to_bytes(prepared_request.body)
                if prepared_request.body is not None else None
            ),
            'headers': prepared_request.headers,
            'url': prepared_request.url,
        }

        for fetch_kwarg in ('connect_timeout', 'timeout', 'tcp_nodelay'):
            if fetch_kwarg in request_params:
                request_for_twisted[fetch_kwarg] = request_params[fetch_kwarg]

        return request_for_twisted


class FidoFutureAdapter(FutureAdapter[T]):
    """
    This is just a wrapper for an EventualResult object from crochet.
    It implements the 'result' method which is needed by our HttpFuture to
    retrieve results.
    """

    timeout_errors = (fido.exceptions.HTTPTimeoutError,)
    connection_errors = (
        fido.exceptions.TCPConnectionError,
        twisted.internet.error.ConnectingCancelledError,
        twisted.internet.error.DNSLookupError,
        twisted.web.client.RequestNotSent,
    )

    def __init__(self, eventual_result):
        # type: (typing.Any) -> None
        self._eventual_result = eventual_result

    def result(self, timeout=None):
        # type: (typing.Optional[float]) -> T
        try:
            return self._eventual_result.wait(timeout=timeout)
        except crochet.TimeoutError:
            self.cancel()
            six.reraise(
                fido.exceptions.HTTPTimeoutError,
                fido.exceptions.HTTPTimeoutError(
                    'Connection was closed by fido after blocking for '
                    'timeout={timeout} seconds waiting for the server to '
                    'send the response'.format(timeout=timeout)
                ),
                sys.exc_info()[2],
            )

    def cancel(self):
        # type: () -> None
        self._eventual_result.cancel()



# -*- coding: utf-8 -*-
import typing
from bravado_core.response import IncomingResponse
from six import with_metaclass

try:
    # ignored type ConnectionError is introduced in python3.3+ (mypy runs as 2.7)
    from builtins import ConnectionError as base_connection_error  # type: ignore
except ImportError:
    # ConnectionError was introduced in python 3.3+
    base_connection_error = OSError


try:
    # ignored type TimeoutError is introduced in python3.3+ (mypy runs as 2.7)
    from builtins import TimeoutError as base_timeout_error  # type: ignore
except ImportError:
    # TimeoutError was introduced in python 3.3+
    base_timeout_error = OSError


if getattr(typing, 'TYPE_CHECKING', False):
    T = typing.TypeVar('T')


# Dictionary of HTTP status codes to exception classes
status_map = {}  # type: typing.MutableMapping[int, typing.Type['HTTPError']]


def _register_exception(exception_class):
    # type: (typing.Type['HTTPError']) -> None
    """Store an HTTP exception class with a status code into a mapping
    of status codes to exception classes.
    :param exception_class: A subclass of HTTPError
    :type exception_class: :class:`HTTPError`
    """
    status_map[exception_class.status_code] = exception_class


class HTTPErrorType(type):
    """A metaclass for registering HTTPError subclasses."""

    def __new__(cls, *args, **kwargs):
        # type: (typing.Type[type], typing.Any, typing.Any) -> type
        new_class = super(HTTPErrorType, cls).__new__(cls, *args, **kwargs)
        if hasattr(new_class, 'status_code'):
            _register_exception(new_class)
        return new_class


class HTTPError(with_metaclass(HTTPErrorType, IOError)):
    """Unified HTTPError used across all http_client implementations.
    """
    status_code = None  # type: int

    def __init__(
        self,
        response,  # type: IncomingResponse
        message=None,   # type: typing.Optional[typing.Text]
        swagger_result=None,  # type: typing.Optional[T]
    ):
        # type: (...) -> None
        """
        :type response: :class:`bravado_core.response.IncomingResponse`
        :param message: Optional string message
        :param swagger_result: If the response for this HTTPError is
            documented in the swagger spec, then this should be the result
            value of the response.
        """
        self.response = response
        self.message = message
        self.swagger_result = swagger_result
        self.status_code = self.response.status_code

    def __str__(self):
        # type: (...) -> str
        # Try to surface the most useful/relevant information available
        # since this is the first thing a developer sees when bad things
        # happen.
        status_and_reason = str(self.response)
        message = ': ' + self.message if self.message else ''
        result = ': {0}'.format(self.swagger_result) \
            if self.swagger_result is not None else ''
        return '{0}{1}{2}'.format(status_and_reason, message, result)


def make_http_exception(
    response,  # type: IncomingResponse
    message=None,  # type: typing.Optional[typing.Text]
    swagger_result=None,  # type: typing.Optional[T]
):
    # type: (...) -> HTTPError
    """
    Return an HTTP exception class  based on the response. If a specific
    class doesn't exist for a particular HTTP status code, a more
    general :class:`HTTPError` class will be returned.
    :type response: :class:`bravado_core.response.IncomingResponse`
    :param message: Optional string message
    :param swagger_result: If the response for this HTTPError is
        documented in the swagger spec, then this should be the result
        value of the response.
    :return: An HTTP exception class that can be raised
    """
    status_code = response.status_code
    exc_class = status_map.get(status_code)

    if exc_class is None:
        exception_families = {
            300: HTTPRedirection,
            400: HTTPClientError,
            500: HTTPServerError,
        }
        exc_class = exception_families.get(
            (status_code // 100) * 100,
        ) or HTTPError

    return exc_class(response, message=message, swagger_result=swagger_result)


class HTTPRedirection(HTTPError):
    """3xx responses."""


class HTTPClientError(HTTPError):
    """4xx responses."""


class HTTPServerError(HTTPError):
    """5xx responses."""


# The follow are based on the HTTP Status Code Registry at
# http://www.iana.org/assignments/http-status-codes/http-status-codes.xhtml
class HTTPMultipleChoices(HTTPRedirection):
    """HTTP/300 - Multiple Choices"""
    status_code = 300


class HTTPMovedPermanently(HTTPRedirection):
    """HTTP/301 - Moved Permanently"""
    status_code = 301


class HTTPFound(HTTPRedirection):
    """HTTP/302 - Found"""
    status_code = 302


class HTTPSeeOther(HTTPRedirection):
    """HTTP/303 - See Other"""
    status_code = 303


class HTTPNotModified(HTTPRedirection):
    """HTTP/304 - Not Modified"""
    status_code = 304


class HTTPUseProxy(HTTPRedirection):
    """HTTP/305 - Use Proxy"""
    status_code = 305


class HTTPTemporaryRedirect(HTTPRedirection):
    """HTTP/307 - Temporary Redirect"""
    status_code = 307


class HTTPPermanentRedirect(HTTPRedirection):
    """HTTP/308 - Permanent Redirect"""
    status_code = 308


class HTTPBadRequest(HTTPClientError):
    """HTTP/400 - Bad Request"""
    status_code = 400


class HTTPUnauthorized(HTTPClientError):
    """HTTP/401 - Unauthorized"""
    status_code = 401


class HTTPPaymentRequired(HTTPClientError):
    """HTTP/402 - Payment Required"""
    status_code = 402


class HTTPForbidden(HTTPClientError):
    """HTTP/403 - Forbidden"""
    status_code = 403


class HTTPNotFound(HTTPClientError):
    """HTTP/404 - Not Found"""
    status_code = 404


class HTTPMethodNotAllowed(HTTPClientError):
    """HTTP/405 - Method Not Allowed"""
    status_code = 405


class HTTPNotAcceptable(HTTPClientError):
    """HTTP/406 - Not Acceptable"""
    status_code = 406


class HTTPProxyAuthenticationRequired(HTTPClientError):
    """HTTP/407 - Proxy Authentication Required"""
    status_code = 407


class HTTPRequestTimeout(HTTPClientError):
    """HTTP/408 - Request Timeout"""
    status_code = 408


class HTTPConflict(HTTPClientError):
    """HTTP/409 - Conflict"""
    status_code = 409


class HTTPGone(HTTPClientError):
    """HTTP/410 - Gone"""
    status_code = 410


class HTTPLengthRequired(HTTPClientError):
    """HTTP/411 - Length Required"""
    status_code = 411


class HTTPPreconditionFailed(HTTPClientError):
    """HTTP/412 - Precondition Failed"""
    status_code = 412


class HTTPPayloadTooLarge(HTTPClientError):
    """HTTP/413 - Payload Too Large"""
    status_code = 413


class HTTPURITooLong(HTTPClientError):
    """HTTP/414 - URI Too Long"""
    status_code = 414


class HTTPUnsupportedMediaType(HTTPClientError):
    """HTTP/415 - Unsupported Media Type"""
    status_code = 415


class HTTPRangeNotSatisfiable(HTTPClientError):
    """HTTP/416 - Range Not Satisfiable"""
    status_code = 416


class HTTPExpectationFailed(HTTPClientError):
    """HTTP/417 - Expectation Failed"""
    status_code = 417


class HTTPMisdirectedRequest(HTTPClientError):
    """HTTP/421 - Misdirected Request"""
    status_code = 421


class HTTPUnprocessableEntity(HTTPClientError):
    """HTTP/422 - Unprocessable Entity"""
    status_code = 422


class HTTPLocked(HTTPClientError):
    """HTTP/423 - Locked"""
    status_code = 423


class HTTPFailedDependency(HTTPClientError):
    """HTTP/424 - Failed Dependency"""
    status_code = 424


class HTTPUpgradeRequired(HTTPClientError):
    """HTTP/426 - Upgrade Required"""
    status_code = 426


class HTTPPreconditionRequired(HTTPClientError):
    """HTTP/428 - Precondition Required"""
    status_code = 428


class HTTPTooManyRequests(HTTPClientError):
    """HTTP/429 - Too Many Requests"""
    status_code = 429


class HTTPRequestHeaderFieldsTooLarge(HTTPClientError):
    """HTTP/431 - Request Header Fields Too Large"""
    status_code = 431


class HTTPUnavailableForLegalReasons(HTTPClientError):
    """HTTP/451 - Unavailable For Legal Reasons"""
    status_code = 451


class HTTPInternalServerError(HTTPServerError):
    """HTTP/500 - Internal Server Error"""
    status_code = 500


class HTTPNotImplemented(HTTPServerError):
    """HTTP/501 - Not Implemented"""
    status_code = 501


class HTTPBadGateway(HTTPServerError):
    """HTTP/502 - Bad Gateway"""
    status_code = 502


class HTTPServiceUnavailable(HTTPServerError):
    """HTTP/503 - Service Unavailable"""
    status_code = 503


class HTTPGatewayTimeout(HTTPServerError):
    """HTTP/504 - Gateway Timeout"""
    status_code = 504


class HTTPHTTPVersionNotSupported(HTTPServerError):
    """HTTP/505 - HTTP Version Not Supported"""
    status_code = 505


class HTTPVariantAlsoNegotiates(HTTPServerError):
    """HTTP/506 - Variant Also Negotiates"""
    status_code = 506


class HTTPInsufficientStorage(HTTPServerError):
    """HTTP/507 - Insufficient Storage"""
    status_code = 507


class HTTPLoopDetected(HTTPServerError):
    """HTTP/508 - Loop Detected"""
    status_code = 508


class HTTPNotExtended(HTTPServerError):
    """HTTP/510 - Not Extended"""
    status_code = 510


class HTTPNetworkAuthenticationRequired(HTTPServerError):
    """HTTP/511 - Network Authentication Required"""
    status_code = 511


class BravadoTimeoutError(base_timeout_error):
    pass


class BravadoConnectionError(base_connection_error):
    pass


class ForcedFallbackResultError(Exception):
    """This exception will be handled if the option to force returning a fallback result
     is used."""



# -*- coding: utf-8 -*-
import logging
from importlib import import_module

import typing
from bravado_core.operation import Operation
from bravado_core.response import IncomingResponse

from bravado.response import BravadoResponseMetadata

try:
    from typing import Type
except ImportError:
    # Python 3.5.0 / 3.5.1
    from typing_extensions import Type


log = logging.getLogger(__name__)


CONFIG_DEFAULTS = {
    # See the constructor of :class:`bravado.http_future.HttpFuture` for an
    # in depth explanation of what this means.
    'also_return_response': False,
    # Kill switch to disable returning fallback results even if provided.
    'disable_fallback_results': False,
    'response_metadata_class': 'bravado.response.BravadoResponseMetadata',
}


BravadoConfig = typing.NamedTuple(
    'BravadoConfig',
    (
        ('also_return_response', bool),
        ('disable_fallback_results', bool),
        ('response_metadata_class', Type[BravadoResponseMetadata]),
    ),
)


def bravado_config_from_config_dict(config_dict):
    # type: (typing.Mapping[str, typing.Any]) -> 'BravadoConfig'
    if config_dict is None:
        config_dict = {}
    bravado_config = {key: value for key, value in config_dict.items() if key in BravadoConfig._fields}
    bravado_config = dict(CONFIG_DEFAULTS, **bravado_config)
    bravado_config['response_metadata_class'] = _get_response_metadata_class(
        bravado_config['response_metadata_class'],
    )
    return BravadoConfig(
        **bravado_config
    )


class RequestConfig(object):

    also_return_response = False  # type: bool
    force_fallback_result = False  # type: bool

    # List of callbacks that are executed after the incoming response has been
    # validated and the swagger_result has been unmarshalled.
    #
    # The callback should expect two arguments:
    #   param : incoming_response
    #   type  : subclass of class:`bravado_core.response.IncomingResponse`
    #   param : operation
    #   type  : class:`bravado_core.operation.Operation`
    response_callbacks = []  # type: typing.List[typing.Callable[[IncomingResponse, Operation], None]]

    # options used to construct the request params
    connect_timeout = None  # type: typing.Optional[float]
    headers = {}  # type: typing.Mapping[str, str]
    use_msgpack = False  # type: bool
    timeout = None  # type: typing.Optional[float]

    # Extra options passed in that we don't know about
    additional_properties = {}  # type: typing.Mapping[str, typing.Any]

    def __init__(self, request_options, also_return_response_default):
        # type: (typing.Dict[str, typing.Any], bool) -> None
        request_options = request_options.copy()  # don't modify the original object
        self.also_return_response = also_return_response_default

        for key in list(request_options.keys()):
            if hasattr(self, key):
                setattr(self, key, request_options.pop(key))

        self.additional_properties = request_options


def _get_response_metadata_class(fully_qualified_class_str):
    # type: (str) -> typing.Type[BravadoResponseMetadata]

    class_to_import = _import_class(fully_qualified_class_str)
    if not class_to_import:
        return BravadoResponseMetadata

    if issubclass(class_to_import, BravadoResponseMetadata):
        return class_to_import

    log.warning(
        'bravado configuration error: the metadata class \'%s\' does not extend '
        'BravadoResponseMetadata. Using default class instead.',
        fully_qualified_class_str,
    )
    return BravadoResponseMetadata


def _import_class(fully_qualified_class_str):
    # type: (str) -> typing.Optional[typing.Type]
    try:
        module_name, class_name = fully_qualified_class_str.rsplit('.', 1)
        return getattr(import_module(module_name), class_name)
    except (ImportError, AttributeError, ValueError) as e:
        log.warning(
            'Error while importing \'%s\': %s: %s',
            fully_qualified_class_str,
            type(e),
            str(e),
        )
        return None



# -*- coding: utf-8 -*-
import monotonic
import typing
from bravado_core.response import IncomingResponse

if getattr(typing, 'TYPE_CHECKING', False):  # Needed to avoid cyclic import.
    from bravado.config import RequestConfig


T = typing.TypeVar('T')


class BravadoResponse(typing.Generic[T]):
    """Bravado response object containing the swagger result as well as response metadata.

    :ivar result: Swagger result from the server
    :ivar BravadoResponseMetadata metadata: metadata for this response including HTTP response
    """

    def __init__(
        self,
        result,  # type: typing.Optional[T]
        metadata,  # type: 'BravadoResponseMetadata[T]'
    ):
        # type: (...) -> None
        self.result = result
        self.metadata = metadata

    @property
    def incoming_response(self):
        # type: () -> IncomingResponse
        return self.metadata.incoming_response


class BravadoResponseMetadata(typing.Generic[T]):
    """HTTP response metadata.

    NOTE: The `elapsed_time` attribute might be slightly lower than the actual time spent since calling
    the operation object, as we only start measuring once the call to `HTTPClient.request` returns.
    Nevertheless, it should be accurate enough for logging and debugging, i.e. determining what went
    on and how much time was spent waiting for the response.

    :ivar float start_time: monotonic timestamp at which the future was created
    :ivar float request_end_time: monotonic timestamp at which we received the HTTP response
    :ivar float processing_end_time: monotonic timestamp at which processing the response ended
    :ivar tuple handled_exception_info: 3-tuple of exception class, exception instance and string
        representation of the traceback in case an exception was caught during request processing.
    """

    def __init__(
        self,
        incoming_response,  # type: IncomingResponse
        swagger_result,  # type: typing.Optional[T]
        start_time,  # type: float
        request_end_time,  # type: float
        handled_exception_info,  # type: typing.List[typing.Union[typing.Type[BaseException], BaseException, typing.Text]]  # noqa
        request_config,  # type: RequestConfig
    ):
        # type: (...) -> None
        """
        :param incoming_response: a subclass of bravado_core.response.IncomingResponse.
        :param swagger_result: the unmarshalled result that is being returned to the user.
        :param start_time: monotonic timestamp indicating when the HTTP future was created. Depending on the
            internal operation of the HTTP client used, this is either before the HTTP request was initiated
            (default client) or right after the HTTP request was sent (e.g. bravado-asyncio / fido).
        :param request_end_time: monotonic timestamp indicating when we received the incoming response,
            excluding unmarshalling, validation or potential fallback result processing.
        :param handled_exception_info: sys.exc_info() data if an exception was caught and handled as
            part of a fallback response; note that the third element in the list is a string representation
            of the traceback, not a traceback object.
        :param RequestConfig request_config: namedtuple containing the request options that were used
            for making this request.
        """
        self._incoming_response = incoming_response
        self.start_time = start_time
        self.request_end_time = request_end_time
        self.processing_end_time = monotonic.monotonic()
        self.handled_exception_info = handled_exception_info
        self.request_config = request_config

        # we expose the result to the user through the BravadoResponse object;
        # we're passing it in to this object in case custom implementations need it
        self._swagger_result = swagger_result

    @property
    def incoming_response(self):
        # type: () -> IncomingResponse
        if not self._incoming_response:
            raise ValueError('No incoming_response present')
        return self._incoming_response

    @property
    def status_code(self):
        # type: () -> int
        return self.incoming_response.status_code

    @property
    def headers(self):
        # type: () -> typing.Mapping[typing.Text, typing.Text]
        return self.incoming_response.headers

    @property
    def is_fallback_result(self):
        # type: () -> bool
        return self.handled_exception_info is not None

    @property
    def request_elapsed_time(self):
        # type: () -> float
        return self.request_end_time - self.start_time

    @property
    def elapsed_time(self):
        # type: () -> float
        return self.processing_end_time - self.start_time



# -*- coding: utf-8 -*-
from bravado_core.response import IncomingResponse

from bravado.exception import BravadoTimeoutError
from bravado.http_future import FALLBACK_EXCEPTIONS
from bravado.http_future import SENTINEL
from bravado.response import BravadoResponseMetadata


class IncomingResponseMock(IncomingResponse):
    def __init__(self, status_code, **kwargs):
        self.headers = {}
        self.status_code = status_code
        for name, value in kwargs.items():
            setattr(self, name, value)


class BravadoResponseMock(object):
    """Class that behaves like the :meth:`.HttpFuture.response` method as well as a :class:`.BravadoResponse`.
    Please check the documentation for further information.
    """

    def __init__(self, result, metadata=None):
        self._result = result
        if metadata:
            self._metadata = metadata
        else:
            self._metadata = BravadoResponseMetadata(
                incoming_response=IncomingResponseMock(status_code=200),
                swagger_result=self._result,
                start_time=1528733800,
                request_end_time=1528733801,
                handled_exception_info=None,
                request_config=None,
            )

    def __call__(self, timeout=None, fallback_result=SENTINEL, exceptions_to_catch=FALLBACK_EXCEPTIONS):
        return self

    @property
    def result(self):
        return self._result

    @property
    def metadata(self):
        return self._metadata


class FallbackResultBravadoResponseMock(object):
    """Class that behaves like the :meth:`.HttpFuture.response` method as well as a :class:`.BravadoResponse`.
    It will always call the ``fallback_result`` callback that's passed to the ``response()`` method.
    Please check the documentation for further information.
    """

    def __init__(self, exception=BravadoTimeoutError(), metadata=None):
        self._exception = exception
        if metadata:
            self._metadata = metadata
        else:
            self._metadata = BravadoResponseMetadata(
                incoming_response=None,
                swagger_result=None,  # we're going to set it later
                start_time=1528733800,
                request_end_time=1528733801,
                handled_exception_info=(self._exception.__class__, self._exception, 'Traceback'),
                request_config=None,
            )

    def __call__(self, timeout=None, fallback_result=SENTINEL, exceptions_to_catch=FALLBACK_EXCEPTIONS):
        assert fallback_result is not SENTINEL, 'You\'re using FallbackResultBravadoResponseMock without a callable ' \
            'fallback_result. Either provide a callable or use BravadoResponseMock.'
        self._fallback_result = fallback_result(self._exception) if callable(fallback_result) else fallback_result
        self._metadata._swagger_result = self._fallback_result
        return self

    @property
    def result(self):
        return self._fallback_result

    @property
    def metadata(self):
        return self._metadata



# -*- coding: utf-8 -*-
import logging
import sys
import traceback
from functools import wraps
from itertools import chain

import monotonic
import six
import typing
from bravado_core.content_type import APP_JSON
from bravado_core.content_type import APP_MSGPACK
from bravado_core.exception import MatchingResponseNotFound
from bravado_core.operation import Operation
from bravado_core.response import get_response_spec
from bravado_core.response import IncomingResponse
from bravado_core.unmarshal import unmarshal_schema_object
from bravado_core.validate import validate_schema_object
from msgpack import unpackb

from bravado.config import bravado_config_from_config_dict
from bravado.config import BravadoConfig
from bravado.config import CONFIG_DEFAULTS
from bravado.config import RequestConfig
from bravado.exception import BravadoConnectionError
from bravado.exception import BravadoTimeoutError
from bravado.exception import ForcedFallbackResultError
from bravado.exception import HTTPServerError
from bravado.exception import make_http_exception
from bravado.response import BravadoResponse


FuncType = typing.Callable[..., typing.Any]
F = typing.TypeVar('F', bound=FuncType)
T = typing.TypeVar('T')
log = logging.getLogger(__name__)


FALLBACK_EXCEPTIONS = (
    BravadoTimeoutError,
    BravadoConnectionError,
    HTTPServerError,
)


class _SENTINEL(object):
    pass


SENTINEL = _SENTINEL()


class FutureAdapter(typing.Generic[T]):
    """
    Mimics a :class:`concurrent.futures.Future` regardless of which client is
    performing the request, whether it is synchronous or actually asynchronous.

    This adapter must be implemented by all bravado clients such as FidoClient
    or RequestsClient to wrap the object returned by their 'request' method.

    """

    # Make sure to define the timeout errors associated with your http client
    timeout_errors = ()  # type: typing.Tuple[typing.Type[BaseException], ...]
    connection_errors = ()  # type: typing.Tuple[typing.Type[BaseException], ...]

    def _raise_error(self, base_exception_class, class_name_suffix, exception):
        # type: (typing.Type[BaseException], typing.Text, BaseException) -> typing.NoReturn
        error = type(
            '{}{}'.format(self.__class__.__name__, class_name_suffix),
            (exception.__class__, base_exception_class),
            dict(
                # Small hack to allow all exceptions to be generated even if they have parameters in the signature
                exception.__dict__,
                __init__=lambda *args, **kwargs: None,
            ),
        )()

        six.reraise(
            error.__class__,
            error,
            sys.exc_info()[2],
        )

    def _raise_timeout_error(self, exception):
        # type: (BaseException) -> typing.NoReturn
        self._raise_error(BravadoTimeoutError, 'Timeout', exception)

    def _raise_connection_error(self, exception):
        # type: (BaseException) -> typing.NoReturn
        self._raise_error(BravadoConnectionError, 'ConnectionError', exception)

    def result(self, timeout=None):
        # type: (typing.Optional[float]) -> T
        """
        Must implement a result method which blocks on result retrieval.

        :param timeout: maximum time to wait on result retrieval. Defaults to
            None which means blocking undefinitely.
        """
        raise NotImplementedError(
            "FutureAdapter must implement 'result' method"
        )

    def cancel(self):
        # type: () -> None
        """
        Must implement a cancel method which terminates any ongoing network
        request.
        """
        log.warning('The FutureAdapter class of the HTTP client does not implement cancel(), ignoring the call.')


def reraise_errors(func):
    # type: (F) -> F

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # type: (typing.Any, typing.Any, typing.Any) -> typing.Any
        timeout_errors = tuple(self.future.timeout_errors or ())
        connection_errors = tuple(self.future.connection_errors or ())

        try:
            return func(self, *args, **kwargs)
        except timeout_errors as exception:
            self.future._raise_timeout_error(exception)
        except connection_errors as exception:
            self.future._raise_connection_error(exception)

    return typing.cast(F, wrapper)


class HttpFuture(typing.Generic[T]):
    """Wrapper for a :class:`FutureAdapter` that returns an HTTP response.

    :param future: The future object to wrap.
    :type future: :class: `FutureAdapter`
    :param response_adapter: Adapter type which exposes the innards of the HTTP
        response in a non-http client specific way.
    :type response_adapter: type that is a subclass of
        :class:`bravado_core.response.IncomingResponse`.
    :param RequestConfig request_config: See :class:`bravado.config.RequestConfig` and
        :data:`bravado.client.REQUEST_OPTIONS_DEFAULTS`
    """

    def __init__(
        self,
        future,  # type: FutureAdapter
        response_adapter,  # type: typing.Type[IncomingResponse]
        operation=None,  # type: typing.Optional[Operation]
        request_config=None,  # type: typing.Optional[RequestConfig]
    ):
        # type: (...) -> None
        self._start_time = monotonic.monotonic()
        self.future = future
        self.response_adapter = response_adapter
        self.operation = operation
        self.request_config = request_config or RequestConfig(
            {},
            also_return_response_default=False,
        )

    @property
    def _bravado_config(self):
        # type: () -> BravadoConfig
        if self.operation:
            return self.operation.swagger_spec.config['bravado']
        else:
            return bravado_config_from_config_dict(CONFIG_DEFAULTS)

    def response(
        self,
        timeout=None,  # type: typing.Optional[float]
        fallback_result=SENTINEL,  # type: typing.Union[_SENTINEL, T, typing.Callable[[BaseException], T]]  # noqa
        exceptions_to_catch=FALLBACK_EXCEPTIONS,  # type: typing.Tuple[typing.Type[BaseException], ...]
    ):
        # type: (...) -> BravadoResponse[T]
        """Blocking call to wait for the HTTP response.

        :param timeout: Number of seconds to wait for a response. Defaults to
            None which means wait indefinitely.
        :type timeout: float
        :param fallback_result: either the swagger result or a callable that accepts an exception as argument
            and returns the swagger result to use in case of errors
        :param exceptions_to_catch: Exception classes to catch and call `fallback_result`
            with. Has no effect if `fallback_result` is not provided. By default, `fallback_result`
            will be called for read timeout and server errors (HTTP 5XX).
        :return: A BravadoResponse instance containing the swagger result and response metadata.
        """
        incoming_response = None
        exc_info = []  # type: typing.List[typing.Union[typing.Type[BaseException], BaseException, typing.Text]]
        request_end_time = None
        if self.request_config.force_fallback_result:
            exceptions_to_catch = tuple(chain(exceptions_to_catch, (ForcedFallbackResultError,)))

        try:
            incoming_response = self._get_incoming_response(timeout)
            request_end_time = monotonic.monotonic()

            swagger_result = self._get_swagger_result(incoming_response)

            if self.operation is None and incoming_response.status_code >= 300:
                raise make_http_exception(response=incoming_response)

            # Trigger fallback_result if the option is set
            if fallback_result is not SENTINEL and self.request_config.force_fallback_result:
                if self._bravado_config.disable_fallback_results:
                    log.warning(
                        'force_fallback_result set in request options and disable_fallback_results '
                        'set in client config; not using fallback result.'
                    )
                else:
                    # raise an exception to trigger fallback result handling
                    raise ForcedFallbackResultError()

        except exceptions_to_catch as e:
            if request_end_time is None:
                request_end_time = monotonic.monotonic()
            exc_info = []
            # the return values of exc_info are annotated as Optional, but we know they are set in this case.
            # additionally, we can't use a cast() since that caused a runtime exception on some older versions
            # of Python 3.5.
            exc_info.extend(sys.exc_info()[:2])  # type: ignore
            # the Python 2 documentation states that we shouldn't assign the traceback to a local variable,
            # as that would cause a circular reference. We'll store a string representation of the traceback
            # instead.
            exc_info.append(traceback.format_exc())

            if (
                fallback_result is not SENTINEL and
                self.operation and
                not self.operation.swagger_spec.config['bravado'].disable_fallback_results
            ):
                if callable(fallback_result):
                    swagger_result = fallback_result(e)
                else:
                    swagger_result = typing.cast(T, fallback_result)
            else:
                six.reraise(*sys.exc_info())

        metadata_class = self._bravado_config.response_metadata_class
        response_metadata = metadata_class(
            incoming_response=incoming_response,
            swagger_result=swagger_result,
            start_time=self._start_time,
            request_end_time=request_end_time,
            handled_exception_info=exc_info,
            request_config=self.request_config,
        )
        return BravadoResponse(
            result=swagger_result,
            metadata=response_metadata,
        )

    def result(
        self,
            timeout=None,  # type: typing.Optional[float]
    ):
        # type: (...) -> typing.Union[T, IncomingResponse, typing.Tuple[T, IncomingResponse]]
        """DEPRECATED: please use the `response()` method instead.

        Blocking call to wait for and return the unmarshalled swagger result.

        :param timeout: Number of seconds to wait for a response. Defaults to
            None which means wait indefinitely.
        :type timeout: float
        :return: Depends on the value of also_return_response sent in
            to the constructor.
        """
        incoming_response = self._get_incoming_response(timeout)
        swagger_result = self._get_swagger_result(incoming_response)

        if self.operation is not None:
            if self.request_config.also_return_response:
                return swagger_result, incoming_response
            return swagger_result

        if 200 <= incoming_response.status_code < 300:
            return incoming_response

        raise make_http_exception(response=incoming_response)

    def cancel(self):
        # type: () -> None
        return self.future.cancel()

    @reraise_errors
    def _get_incoming_response(self, timeout=None):
        # type: (typing.Optional[float]) -> IncomingResponse
        inner_response = self.future.result(timeout=timeout)
        incoming_response = self.response_adapter(inner_response)
        return incoming_response

    @reraise_errors  # unmarshal_response_inner calls response.json(), which might raise errors
    def _get_swagger_result(self, incoming_response):
        # type: (IncomingResponse) -> typing.Optional[T]
        swagger_result = None
        if self.operation is not None:
            unmarshal_response(
                incoming_response,
                self.operation,
                self.request_config.response_callbacks,
            )
            swagger_result = typing.cast(T, incoming_response.swagger_result)

        return swagger_result


def unmarshal_response(
    incoming_response,  # type: IncomingResponse
    operation,  # type: Operation
    response_callbacks=None,  # type: typing.Optional[typing.List[typing.Callable[[typing.Any, typing.Any], None]]]
):
    # type: (...) -> None
    """So the http_client is finished with its part of processing the response.
    This hands the response over to bravado_core for validation and
    unmarshalling and then runs any response callbacks. On success, the
    swagger_result is available as ``incoming_response.swagger_result``.
    :type incoming_response: :class:`bravado_core.response.IncomingResponse`
    :type operation: :class:`bravado_core.operation.Operation`
    :type response_callbacks: list of callable. See
        bravado_core.client.REQUEST_OPTIONS_DEFAULTS.
    :raises: HTTPError
        - On 5XX status code, the HTTPError has minimal information.
        - On non-2XX status code with no matching response, the HTTPError
            contains a detailed error message.
        - On non-2XX status code with a matching response, the HTTPError
            contains the return value.
    """
    response_callbacks = response_callbacks or []

    try:
        raise_on_unexpected(incoming_response)
        incoming_response.swagger_result = unmarshal_response_inner(
            response=incoming_response,
            op=operation,
        )
    except MatchingResponseNotFound as e:
        exception = make_http_exception(
            response=incoming_response,
            message=str(e)
        )
        six.reraise(
            type(exception),
            exception,
            sys.exc_info()[2])
    finally:
        # Always run the callbacks regardless of success/failure
        for response_callback in response_callbacks:
            response_callback(incoming_response, operation)

    raise_on_expected(incoming_response)


def unmarshal_response_inner(
    response,  # type: IncomingResponse
    op,  # type: Operation
):
    # type: (...) -> typing.Optional[T]
    """
    Unmarshal incoming http response into a value based on the
    response specification.
    :type response: :class:`bravado_core.response.IncomingResponse`
    :type op: :class:`bravado_core.operation.Operation`
    :returns: value where type(value) matches response_spec['schema']['type']
        if it exists, None otherwise.
    """
    deref = op.swagger_spec.deref
    response_spec = get_response_spec(status_code=response.status_code, op=op)

    if 'schema' not in response_spec:
        return None

    content_type = response.headers.get('content-type', '').lower()

    if content_type.startswith(APP_JSON) or content_type.startswith(APP_MSGPACK):
        content_spec = deref(response_spec['schema'])
        if content_type.startswith(APP_JSON):
            content_value = response.json()
        else:
            content_value = unpackb(response.raw_bytes, encoding='utf-8')

        if op.swagger_spec.config.get('validate_responses', False):
            validate_schema_object(op.swagger_spec, content_spec, content_value)

        return unmarshal_schema_object(
            swagger_spec=op.swagger_spec,
            schema_object_spec=content_spec,
            value=content_value,
        )

    # TODO: Non-json response contents
    return response.text


def raise_on_unexpected(http_response):
    # type: (IncomingResponse) -> None
    """Raise an HTTPError if the response is 5XX.

    :param http_response: :class:`bravado_core.response.IncomingResponse`
    :raises: HTTPError
    """
    if 500 <= http_response.status_code <= 599:
        raise make_http_exception(response=http_response)


def raise_on_expected(http_response):
    # type: (IncomingResponse) -> None
    """Raise an HTTPError if the response is non-2XX and matches a response
    in the swagger spec.

    :param http_response: :class:`bravado_core.response.IncomingResponse`
    :raises: HTTPError
    """
    if not 200 <= http_response.status_code < 300:
        raise make_http_exception(
            response=http_response,
            swagger_result=http_response.swagger_result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2013, Digium, Inc.
# Copyright (c) 2014-2016, Yelp, Inc.
import os

from setuptools import find_packages
from setuptools import setup

import bravado

setup(
    name='bravado',
    version=bravado.version,
    license='BSD 3-Clause License',
    description='Library for accessing Swagger-enabled API\'s',
    long_description=open(os.path.join(os.path.dirname(__file__),
                                       'README.rst')).read(),
    author='Digium, Inc. and Yelp, Inc.',
    author_email='opensource+bravado@yelp.com',
    url='https://github.com/Yelp/bravado',
    packages=find_packages(include=['bravado*']),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    package_data={
        'bravado': ['py.typed'],
    },
    install_requires=[
        'bravado-core >= 5.0.1',
        'msgpack-python',
        'python-dateutil',
        'pyyaml',
        'requests >= 2.4',
        'six',
        'simplejson',
        'monotonic',
    ],
    extras_require={
        'fido': ['fido >= 4.2.1'],
        ':python_version<"3.5"': ['typing'],
        # only needed for Python 3.5.0 and 3.5.1, but =="3.5.0" evaluates
        # to True for all Python 3.5 versions apparently
        ':python_version=="3.5"': ['typing_extensions'],
        'integration-tests': [
            'bottle',
            'ephemeral_port_reserve',
            'pytest',
        ]
    },
)



# -*- coding: utf-8 -*-
import json
import time
from multiprocessing import Process

import bottle
import ephemeral_port_reserve
import pytest
import requests
import requests.exceptions
import typing
from bravado_core.content_type import APP_MSGPACK
from msgpack import packb
from msgpack import unpackb

from bravado.client import SwaggerClient
from bravado.exception import BravadoConnectionError
from bravado.exception import BravadoTimeoutError
from bravado.http_client import HttpClient
from bravado.http_future import FutureAdapter
from bravado.swagger_model import Loader


def _class_fqn(clz):
    return '{t.__module__}.{t.__name__}'.format(t=clz)


ROUTE_1_RESPONSE = b'HEY BUDDY'
ROUTE_2_RESPONSE = b'BYE BUDDY'
API_RESPONSE = {'answer': 42}
SWAGGER_SPEC_DICT = {
    'swagger': '2.0',
    'info': {'version': '1.0.0', 'title': 'Integration tests'},
    'definitions': {
        'api_response': {
            'properties': {
                'answer': {
                    'type': 'integer'
                },
            },
            'required': ['answer'],
            'type': 'object',
            'x-model': 'api_response',
            'title': 'api_response',
        }
    },
    'basePath': '/',
    'paths': {
        '/json': {
            'get': {
                'produces': ['application/json'],
                'responses': {
                    '200': {
                        'description': 'HTTP/200',
                        'schema': {'$ref': '#/definitions/api_response'},
                    },
                },
            },
        },
        '/json_or_msgpack': {
            'get': {
                'produces': [
                    'application/msgpack',
                    'application/json'
                ],
                'responses': {
                    '200': {
                        'description': 'HTTP/200',
                        'schema': {'$ref': '#/definitions/api_response'},
                    }
                }
            }
        },
        '/echo': {
            'get': {
                'produces': ['application/json'],
                'parameters': [
                    {
                        'in': 'query',
                        'name': 'message',
                        'type': 'string',
                        'required': True,
                    }
                ],
                'responses': {
                    '200': {
                        'description': 'HTTP/200',
                        'schema': {
                            'type': 'object',
                            'properties': {
                                'message': {
                                    'type': 'string',
                                },
                            },
                            'required': ['message'],
                        },
                    },
                },
            },
        },
        '/char_test/{special}': {
            'get': {
                'operationId': 'get_char_test',
                'produces': ['application/json'],
                'parameters': [
                    {
                        'in': 'path',
                        'name': 'special',
                        'type': 'string',
                        'required': True,
                    }
                ],
                'responses': {
                    '200': {
                        'description': 'HTTP/200',
                        'schema': {'$ref': '#/definitions/api_response'},
                    },
                },
            },
        },
        '/sanitize-test': {
            'get': {
                'operationId': 'get_sanitized_param',
                'produces': ['application/json'],
                'parameters': [
                    {
                        'in': 'header',
                        'name': 'X-User-Id',
                        'type': 'string',
                        'required': True,
                    }
                ],
                'responses': {
                    '200': {
                        'description': 'HTTP/200',
                        'schema': {'$ref': '#/definitions/api_response'},
                    },
                },
            },
        },
        '/sleep': {
            'get': {
                'operationId': 'sleep',
                'produces': ['application/json'],
                'parameters': [
                    {
                        'in': 'query',
                        'name': 'sec',
                        'type': 'number',
                        'format': 'float',
                        'required': True,
                    }
                ],
                'responses': {
                    '200': {
                        'description': 'HTTP/200',
                        'schema': {'$ref': '#/definitions/api_response'},
                    },
                },
            },
        },
    },
}


@bottle.get('/swagger.json')
def swagger_spec():
    return SWAGGER_SPEC_DICT


@bottle.get('/json')
def api_json():
    return API_RESPONSE


@bottle.route('/json_or_msgpack')
def api_json_or_msgpack():
    if bottle.request.headers.get('accept') == APP_MSGPACK:
        bottle.response.content_type = APP_MSGPACK
        return packb(API_RESPONSE)
    else:
        return API_RESPONSE


@bottle.route('/1')
def one():
    return ROUTE_1_RESPONSE


@bottle.route('/2')
def two():
    return ROUTE_2_RESPONSE


@bottle.post('/double')
def double():
    x = bottle.request.params['number']
    return str(int(x) * 2)


@bottle.route('/headers')
def headers():
    return {
        header_name: {
            'type': _class_fqn(type(header_value)),
            'representation': repr(header_value),
        }
        for header_name, header_value in bottle.request.headers.items()
        if header_name.startswith('Header-')
    }


@bottle.get('/echo')
def echo():
    return {'message': bottle.request.params['message']}


@bottle.get('/char_test/spe%ial?')
def char_test():
    return API_RESPONSE


@bottle.get('/sanitize-test')
def sanitize_test():
    if bottle.request.headers.get('X-User-Id') == 'admin':
        return API_RESPONSE
    return bottle.HTTPResponse(status=404)


@bottle.get('/sleep')
def sleep_api():
    sec_to_sleep = float(bottle.request.GET.get('sec', '1'))
    time.sleep(sec_to_sleep)
    return sec_to_sleep


class IntegrationTestingServicesAndClient:
    @pytest.fixture(scope='session')
    def swagger_http_server(self):
        def wait_unit_service_starts(url, max_wait_time=10):
            start = time.time()
            check_url = '{url}/swagger.json'.format(url=url)
            while time.time() < start + max_wait_time:
                try:
                    requests.get(check_url, timeout=1)
                except requests.ConnectionError:
                    time.sleep(0.1)
                else:
                    return

        port = ephemeral_port_reserve.reserve()

        web_service_process = Process(
            target=bottle.run,
            kwargs={'quiet': True, 'host': 'localhost', 'port': port},
        )
        try:
            web_service_process.start()
            server_address = 'http://localhost:{port}'.format(port=port)
            wait_unit_service_starts(server_address, 10)
            yield server_address
        finally:
            web_service_process.terminate()

    @pytest.fixture(scope='session')
    def not_answering_http_server(self):
        yield 'http://localhost:{}'.format(ephemeral_port_reserve.reserve())

    @pytest.fixture
    def swagger_client(self, swagger_http_server):
        return SwaggerClient.from_url(
            spec_url='{server_address}/swagger.json'.format(
                server_address=swagger_http_server),
            http_client=self.http_client,
            config={'use_models': False, 'also_return_response': True}
        )

    @pytest.fixture(params=['result', 'response'])
    def result_getter(self, request):
        if request.param == 'result':
            return lambda future, timeout: future.result(timeout)
        elif request.param == 'response':
            def _response_getter(future, timeout):
                response = future.response(timeout)
                return response.result, response.incoming_response

            return _response_getter

        raise ValueError  # pragma: no cover


class IntegrationTestingFixturesMixin(IntegrationTestingServicesAndClient):
    """
    Generic class to run integration tests with the different HTTP clients definitions
    """

    http_client_type = None  # type: typing.Type[HttpClient]
    http_future_adapter_type = None  # type: typing.Type[FutureAdapter]
    connection_errors_exceptions = None  # type: typing.Set[Exception]

    @classmethod
    def setup_class(cls):
        if cls.http_client_type is None:
            raise RuntimeError(  # pragma: no cover
                'Define http_client_type for {}'.format(cls.__name__)
            )
        if cls.http_future_adapter_type is None:
            raise RuntimeError(  # pragma: no cover
                'Define http_future_adapter_type for {}'.format(cls.__name__)
            )
        if cls.connection_errors_exceptions is None:
            raise RuntimeError(  # pragma: no cover
                'Define connection_errors_exceptions for {}'.format(
                    cls.__name__,
                ),
            )
        cls.http_client = cls.http_client_type()

    @classmethod
    def encode_expected_response(cls, response):
        if isinstance(response, bytes):
            return response.decode('utf-8')
        else:
            return str(response)  # pragma: no cover


class IntegrationTestsBaseClass(IntegrationTestingFixturesMixin):
    def test_fetch_specs(self, swagger_http_server):
        loader = Loader(
            http_client=self.http_client,
            request_headers={'boolean-header': True},
        )
        spec = loader.load_spec('{server_address}/swagger.json'.format(server_address=swagger_http_server))
        assert spec == SWAGGER_SPEC_DICT

    def test_swagger_client_json_response(self, swagger_client, result_getter):
        marshaled_response, raw_response = result_getter(swagger_client.json.get_json(), timeout=1)
        assert marshaled_response == API_RESPONSE
        assert raw_response.raw_bytes == json.dumps(API_RESPONSE).encode('utf-8')

    def test_swagger_client_msgpack_response_without_flag(self, swagger_client, result_getter):
        marshaled_response, raw_response = result_getter(
            swagger_client.json_or_msgpack.get_json_or_msgpack(),
            timeout=1,
        )
        assert marshaled_response == API_RESPONSE
        assert raw_response.raw_bytes == json.dumps(API_RESPONSE).encode('utf-8')

    def test_swagger_client_msgpack_response_with_flag(self, swagger_client, result_getter):
        marshaled_response, raw_response = result_getter(
            swagger_client.json_or_msgpack.get_json_or_msgpack(
                _request_options={
                    'use_msgpack': True,
                },
            ),
            timeout=1,
        )
        assert marshaled_response == API_RESPONSE
        assert raw_response.raw_bytes == packb(API_RESPONSE)

    def test_swagger_client_special_chars_query(self, swagger_client, result_getter):
        message = 'My Me$$age with %pecial characters?"'
        marshaled_response, _ = result_getter(swagger_client.echo.get_echo(message=message), timeout=1)
        assert marshaled_response == {'message': message}

    def test_swagger_client_special_chars_path(self, swagger_client, result_getter):
        marshaled_response, _ = result_getter(swagger_client.char_test.get_char_test(special='spe%ial?'), timeout=1)
        assert marshaled_response == API_RESPONSE

    def test_sanitized_resource_and_param(self, swagger_client, result_getter):
        marshaled_response, _ = result_getter(
            swagger_client.sanitize_test.get_sanitized_param(X_User_Id='admin'),
            timeout=1,
        )
        assert marshaled_response == API_RESPONSE

    def test_unsanitized_resource_and_param(self, swagger_client, result_getter):
        params = {'X-User-Id': 'admin'}
        marshaled_response, _ = result_getter(
            swagger_client._get_resource(
                'sanitize-test',
            ).get_sanitized_param(**params),
            timeout=1,
        )
        assert marshaled_response == API_RESPONSE

    def test_unsanitized_param_as_header(self, swagger_client, result_getter):
        headers = {'X-User-Id': 'admin'}
        marshaled_response, _ = result_getter(swagger_client.sanitize_test.get_sanitized_param(
            _request_options={'headers': headers},
        ), timeout=1)
        assert marshaled_response == API_RESPONSE

    def test_multiple_requests(self, swagger_http_server):
        request_one_params = {
            'method': 'GET',
            'headers': {},
            'url': '{server_address}/1'.format(server_address=swagger_http_server),
            'params': {},
        }

        request_two_params = {
            'method': 'GET',
            'headers': {},
            'url': '{server_address}/2'.format(server_address=swagger_http_server),
            'params': {},
        }

        http_future_1 = self.http_client.request(request_one_params)
        http_future_2 = self.http_client.request(request_two_params)
        resp_one = http_future_1.result(timeout=1)
        resp_two = http_future_2.result(timeout=1)

        assert resp_one.text == self.encode_expected_response(ROUTE_1_RESPONSE)
        assert resp_two.text == self.encode_expected_response(ROUTE_2_RESPONSE)

    def test_post_request(self, swagger_http_server):

        request_args = {
            'method': 'POST',
            'headers': {},
            'url': '{server_address}/double'.format(server_address=swagger_http_server),
            'data': {"number": 3},
        }

        http_future = self.http_client.request(request_args)
        resp = http_future.result(timeout=1)

        assert resp.text == self.encode_expected_response(b'6')

    def test_non_string_headers(self, swagger_http_server):
        headers = {
            'Header-Boolean': True,
            'Header-Integer': 1,
            'Header-Bytes': b'0',
        }
        response = self.http_client.request({
            'method': 'GET',
            'headers': headers,
            'url': '{server_address}/headers'.format(server_address=swagger_http_server),
            'params': {},
        }).result(timeout=1)

        expected_header_representations = {
            'Header-Boolean': repr('True'),
            'Header-Integer': repr('1'),
            'Header-Bytes': repr('0'),
        }
        assert {
            header_name: {
                'type': _class_fqn(str),
                'representation': expected_header_representations[header_name]
            }
            for header_name, header_value in headers.items()
        } == response.json()

    def test_msgpack_support(self, swagger_http_server):
        response = self.http_client.request({
            'method': 'GET',
            'url': '{server_address}/json_or_msgpack'.format(server_address=swagger_http_server),
            'params': {},
            'headers': {
                'Accept': APP_MSGPACK,
            },
        }).result(timeout=1)

        assert response.headers['Content-Type'] == APP_MSGPACK
        assert unpackb(response.raw_bytes, encoding='utf-8') == API_RESPONSE

    def test_timeout_errors_are_thrown_as_BravadoTimeoutError(self, swagger_http_server):
        if not self.http_future_adapter_type.timeout_errors:
            pytest.skip('{} does NOT defines timeout_errors'.format(self.http_future_adapter_type))

        with pytest.raises(BravadoTimeoutError):
            self.http_client.request({
                'method': 'GET',
                'url': '{server_address}/sleep?sec=0.1'.format(server_address=swagger_http_server),
                'params': {},
            }).result(timeout=0.01)

    def test_swagger_client_timeout_errors_are_thrown_as_BravadoTimeoutError(
        self, swagger_client, result_getter,
    ):
        if not self.http_future_adapter_type.timeout_errors:
            pytest.skip('{} does NOT defines timeout_errors'.format(self.http_future_adapter_type))

        with pytest.raises(BravadoTimeoutError):
            result_getter(
                swagger_client.sleep.sleep(sec=0.5),
                timeout=0.1,
            )

    def test_timeout_errors_are_catchable_with_original_exception_types(self, swagger_http_server):
        if not self.http_future_adapter_type.timeout_errors:
            pytest.skip('{} does NOT defines timeout_errors'.format(self.http_future_adapter_type))

        for expected_exception in self.http_future_adapter_type.timeout_errors:
            with pytest.raises(expected_exception) as excinfo:
                self.http_client.request({
                    'method': 'GET',
                    'url': '{server_address}/sleep?sec=0.1'.format(server_address=swagger_http_server),
                    'params': {},
                }).result(timeout=0.01)
            assert isinstance(excinfo.value, BravadoTimeoutError)

    def test_connection_errors_are_thrown_as_BravadoConnectionError(self, not_answering_http_server):
        if not self.http_future_adapter_type.connection_errors:
            pytest.skip('{} does NOT defines connection_errors'.format(self.http_future_adapter_type))

        with pytest.raises(BravadoConnectionError):
            self.http_client.request({
                'method': 'GET',
                'url': '{server_address}/sleep?sec=0.1'.format(server_address=not_answering_http_server),
                'params': {},
                'connect_timeout': 0.001,
                'timeout': 0.01,
            }).result(timeout=1)

    def test_connection_errors_exceptions_contains_all_future_adapter_connection_errors(self):
        assert set(
            type(e) for e in self.connection_errors_exceptions
        ) == set(self.http_future_adapter_type.connection_errors)

    def test_connection_errors_are_catchable_with_original_exception_types(
        self, not_answering_http_server,
    ):
        for expected_exception in self.connection_errors_exceptions:
            with pytest.raises(type(expected_exception)) as excinfo:
                http_future = self.http_client.request({
                    'method': 'GET',
                    'url': not_answering_http_server,
                    'params': {},
                })
                http_future.cancel()

                # Finding a way to force all the http clients to raise
                # the expected exception while sending the real request is hard
                # so we're mocking the future in order to throw the expected
                # exception so we can validate that the exception is catchable
                # with the original exception type too
                def raise_expected_exception(*args, **kwargs):
                    raise expected_exception
                http_future.future.result = raise_expected_exception

                http_future.result(timeout=0.1)

            # check that the raised exception is catchable as BravadoConnectionError too
            assert isinstance(excinfo.value, BravadoConnectionError)

    def test_swagger_client_connection_errors_are_thrown_as_BravadoConnectionError(
        self, not_answering_http_server, swagger_client, result_getter,
    ):
        if not self.http_future_adapter_type.connection_errors:
            pytest.skip('{} does NOT defines connection_errors'.format(self.http_future_adapter_type))

        # override api url to communicate with a non responding http server
        swagger_client.swagger_spec.api_url = not_answering_http_server
        with pytest.raises(BravadoConnectionError):
            result_getter(
                swagger_client.json.get_json(_request_options={
                    'connect_timeout': 0.001,
                    'timeout': 0.01,
                }),
                timeout=0.1,
            )



# -*- coding: utf-8 -*-
import typing
from bravado_core.operation import Operation

from bravado.config import RequestConfig
from bravado.http_future import HttpFuture

APP_FORM = 'application/x-www-form-urlencoded'
MULT_FORM = 'multipart/form-data'


class HttpClient(object):
    """Interface for a minimal HTTP client that can retrieve Swagger specs
    and perform HTTP calls to fulfill a Swagger operation.
    """

    def request(
        self,
        request_params,  # type: typing.MutableMapping[str, typing.Any]
        operation=None,  # type: typing.Optional[Operation]
        request_config=None,  # type: typing.Optional[RequestConfig]
    ):
        # type: (...) -> HttpFuture
        """
        :param request_params: complete request data. e.g. url, method,
            headers, body, params, connect_timeout, timeout, etc.
        :type request_params: dict
        :param operation: operation that this http request is for. Defaults
            to None - in which case, we're obviously just retrieving a Swagger
            Spec.
        :type operation: :class:`bravado_core.operation.Operation`
        :param RequestConfig request_config: Per-request config that is passed to
            :class:`bravado.http_future.HttpFuture`.

        :returns: HTTP Future object
        :rtype: :class: `bravado_core.http_future.HttpFuture`
        """
        raise NotImplementedError(
            u"%s: Method not implemented", self.__class__.__name__)

    def __repr__(self):
        # type: () -> str
        return "{0}()".format(type(self))



# -*- coding: utf-8 -*-
import copy
import logging

import requests
import requests.auth
import requests.exceptions
import six
import typing
from bravado_core.operation import Operation
from bravado_core.response import IncomingResponse
from six import iteritems
from six.moves.urllib import parse as urlparse

from bravado.config import RequestConfig
from bravado.http_client import HttpClient
from bravado.http_future import FutureAdapter
from bravado.http_future import HttpFuture


if getattr(typing, 'TYPE_CHECKING', False):
    T = typing.TypeVar('T')


log = logging.getLogger(__name__)


class Authenticator(object):
    """Authenticates requests.

    :param host: Host to authenticate for.
    """

    def __init__(self, host):
        # type: (str) -> None
        self.host = host

    def __repr__(self):
        # type: () -> str
        return "%s(%s)" % (self.__class__.__name__, self.host)

    def matches(self, url):
        # type: (typing.Text) -> bool
        """Returns true if this authenticator applies to the given url.

        :param url: URL to check.
        :return: True if matches host, port and scheme, False otherwise.
        """
        split = urlparse.urlsplit(url)
        return self.host == split.hostname

    def apply(self, request):
        # type: (requests.Request) -> requests.Request
        """Apply authentication to a request.

        :param request: Request to add authentication information to.
        """
        raise NotImplementedError(u"%s: Method not implemented",
                                  self.__class__.__name__)


# noinspection PyDocstring
class ApiKeyAuthenticator(Authenticator):
    """?api_key authenticator.

    This authenticator adds an API key via query parameter or header.

    :param host: Host to authenticate for.
    :param api_key: API key.
    :param param_name: Query parameter specifying the API key.
    :param param_in: How to send the API key. Can be 'query' or 'header'.
    """

    def __init__(
        self,
        host,  # type: str
        api_key,  # type: typing.Text
        param_name=u'api_key',  # type: typing.Text
        param_in=u'query',  # type: typing.Text
    ):
        # type: (...) -> None
        super(ApiKeyAuthenticator, self).__init__(host)
        self.param_name = param_name
        self.param_in = param_in
        self.api_key = api_key

    def apply(self, request):
        # type: (requests.Request) -> requests.Request
        if self.param_in == 'header':
            request.headers.setdefault(self.param_name, self.api_key)
        else:
            request.params[self.param_name] = self.api_key
        return request


class BasicAuthenticator(Authenticator):
    """HTTP Basic authenticator.

    :param host: Host to authenticate for.
    :param username: Username.
    :param password: Password
    """

    def __init__(
        self,
        host,  # type: str
        username,  # type: typing.Union[bytes, str]
        password,  # type: typing.Union[bytes, str]
    ):
        # type: (...) -> None
        super(BasicAuthenticator, self).__init__(host)
        self.auth = requests.auth.HTTPBasicAuth(username, password)

    def apply(self, request):
        # type: (requests.Request) -> requests.Request
        request.auth = self.auth

        return request


class RequestsClient(HttpClient):
    """Synchronous HTTP client implementation.
    """

    def __init__(self, ssl_verify=True, ssl_cert=None):
        # type: (bool, typing.Any) -> None
        """
        :param ssl_verify: Set to False to disable SSL certificate validation. Provide the path to a
            CA bundle if you need to use a custom one.
        :param ssl_cert: Provide a client-side certificate to use. Either a sequence of strings pointing
            to the certificate (1) and the private key (2), or a string pointing to the combined certificate
            and key.
        """
        self.session = requests.Session()
        self.authenticator = None  # type: typing.Optional[Authenticator]
        self.ssl_verify = ssl_verify
        self.ssl_cert = ssl_cert

    def separate_params(
        self,
        request_params,  # type: typing.MutableMapping[str, typing.Any]
    ):
        # type: (...) -> typing.Tuple[typing.Mapping[str, typing.Any], typing.Mapping[str, typing.Any]]
        """Splits the passed in dict of request_params into two buckets.

        - sanitized_params are valid kwargs for constructing a
          requests.Request(..)
        - misc_options are things like timeouts which can't be communicated
          to the Requests library via the requests.Request(...) constructor.

        :param request_params: kitchen sink of request params. Treated as a
            read-only dict.
        :returns: tuple(sanitized_params, misc_options)
        """
        sanitized_params = copy.copy(request_params)
        misc_options = {
            'ssl_verify': self.ssl_verify,
            'ssl_cert': self.ssl_cert,
        }

        if 'connect_timeout' in sanitized_params:
            misc_options['connect_timeout'] = \
                sanitized_params.pop('connect_timeout')

        if 'timeout' in sanitized_params:
            misc_options['timeout'] = sanitized_params.pop('timeout')

        return sanitized_params, misc_options

    def request(
        self,
        request_params,  # type: typing.MutableMapping[str, typing.Any]
        operation=None,  # type: typing.Optional[Operation]
        request_config=None,  # type: typing.Optional[RequestConfig]
    ):
        # type: (...) -> HttpFuture[T]
        """
        :param request_params: complete request data.
        :type request_params: dict
        :param operation: operation that this http request is for. Defaults
            to None - in which case, we're obviously just retrieving a Swagger
            Spec.
        :type operation: :class:`bravado_core.operation.Operation`
        :param RequestConfig request_config: per-request configuration

        :returns: HTTP Future object
        :rtype: :class: `bravado_core.http_future.HttpFuture`
        """
        sanitized_params, misc_options = self.separate_params(request_params)

        requests_future = RequestsFutureAdapter(
            self.session,
            self.authenticated_request(sanitized_params),
            misc_options,
        )

        return HttpFuture(
            requests_future,
            RequestsResponseAdapter,
            operation,
            request_config,
        )

    def set_basic_auth(
        self,
        host,  # type: str
        username,  # type: typing.Union[bytes, str]
        password,  # type: typing.Union[bytes, str]
    ):
        # type: (...) -> None
        self.authenticator = BasicAuthenticator(
            host=host,
            username=username,
            password=password,
        )

    def set_api_key(
        self,
        host,  # type: str
        api_key,  # type: typing.Text
        param_name=u'api_key',  # type: typing.Text
        param_in=u'query',  # type: typing.Text
    ):
        # type: (...) -> None
        self.authenticator = ApiKeyAuthenticator(
            host=host,
            api_key=api_key,
            param_name=param_name,
            param_in=param_in,
        )

    def authenticated_request(self, request_params):
        # type: (typing.Mapping[str, typing.Any]) -> requests.Request
        return self.apply_authentication(requests.Request(**request_params))

    def apply_authentication(self, request):
        # type: (requests.Request) -> requests.Request
        if self.authenticator and self.authenticator.matches(request.url):
            return self.authenticator.apply(request)

        return request


class RequestsResponseAdapter(IncomingResponse):
    """Wraps a requests.models.Response object to provide a uniform interface
    to the response innards.

    :type requests_lib_response: :class:`requests.models.Response`
    """

    def __init__(self, requests_lib_response):
        # type: (requests.Response) -> None
        self._delegate = requests_lib_response

    @property
    def status_code(self):
        # type: () -> int
        return self._delegate.status_code

    @property
    def text(self):
        # type: () -> typing.Text
        return self._delegate.text

    @property
    def raw_bytes(self):
        # type: () -> bytes
        return self._delegate.content

    @property
    def reason(self):
        # type: () -> typing.Text
        return self._delegate.reason

    @property
    def headers(self):
        # type: () -> typing.Mapping[typing.Text, typing.Text]
        # we don't use typing.cast here since that's broken on Python 3.5.1
        return self._delegate.headers  # type: ignore

    def json(self, **kwargs):
        # type: (typing.Any) -> typing.Mapping[typing.Text, typing.Any]
        return self._delegate.json(**kwargs)


class RequestsFutureAdapter(FutureAdapter):
    """Mimics a :class:`concurrent.futures.Future` for the purposes of making
    HTTP calls with the Requests library in a future-y sort of way.
    """

    timeout_errors = (requests.exceptions.ReadTimeout,)  # type: typing.Tuple[typing.Type[Exception], ...]
    connection_errors = (requests.exceptions.ConnectionError,)  # type: typing.Tuple[typing.Type[Exception], ...]

    def __init__(
        self,
        session,  # type: requests.Session
        request,  # type: requests.Request
        misc_options,  # type: typing.Mapping[str, typing.Any]
    ):
        # type: (...) -> None
        """Kicks API call for Requests client

        :param session: session object to use for making the request
        :param request: dict containing API request parameters
        :param misc_options: misc options to apply when sending a HTTP request.
            e.g. timeout, connect_timeout, etc
        :type misc_options: dict
        """
        self.session = session
        self.request = request
        self.misc_options = misc_options

    def build_timeout(
        self,
        result_timeout,  # type: typing.Optional[float]
    ):
        # type: (...) -> typing.Union[typing.Optional[float], typing.Tuple[typing.Optional[float], typing.Optional[float]]]  # noqa
        """
        Build the appropriate timeout object to pass to `session.send(...)`
        based on connect_timeout, the timeout passed to the service call, and
        the timeout passed to the result call.

        :param result_timeout: timeout that was passed into `future.result(..)`
        :return: timeout
        :rtype: float or tuple(connect_timeout, timeout)
        """
        # The API provides two ways to pass a timeout :( We're stuck
        # dealing with it until we're ready to make a non-backwards
        # compatible change.
        #
        #  - If both timeouts are the same, no problem
        #  - If either has a non-None value, use the non-None value
        #  - If both have a non-None value, use the greater of the two
        timeout = None
        has_service_timeout = 'timeout' in self.misc_options
        service_timeout = self.misc_options.get('timeout')

        if not has_service_timeout:
            timeout = result_timeout
        elif service_timeout == result_timeout:
            timeout = service_timeout
        else:
            if service_timeout is None:
                timeout = result_timeout
            elif result_timeout is None:
                timeout = service_timeout
            else:
                timeout = max(service_timeout, result_timeout)
            log.warning(
                "Two different timeouts have been passed: "
                "_request_options['timeout'] = %s and "
                "future.result(timeout=%s). Using timeout of %s.",
                service_timeout, result_timeout, timeout,
            )

        # Requests is weird in that if you want to specify a connect_timeout
        # and idle timeout, then the timeout is passed as a tuple
        if 'connect_timeout' in self.misc_options:
            return self.misc_options['connect_timeout'], timeout
        return timeout

    def result(self, timeout=None):
        # type: (typing.Optional[float]) -> T
        """Blocking call to wait for API response

        :param timeout: timeout in seconds to wait for response. Defaults to
            None to wait indefinitely.
        :type timeout: float
        :return: raw response from the server
        :rtype: dict
        """
        request = self.request

        # Ensure that all the headers are converted to strings.
        # This is need to workaround https://github.com/requests/requests/issues/3491
        request.headers = {
            k: str(v) if not isinstance(v, six.binary_type) else v
            for k, v in iteritems(request.headers)
        }

        prepared_request = self.session.prepare_request(request)
        settings = self.session.merge_environment_settings(
            prepared_request.url,
            proxies={},
            stream=None,
            verify=self.misc_options['ssl_verify'],
            cert=self.misc_options['ssl_cert'],
        )
        response = self.session.send(
            prepared_request,
            timeout=self.build_timeout(timeout),
            **settings
        )
        return response

    def cancel(self):
        # type: () -> None
        pass
